@@ -1,4 +1,5 @@
 import radcad.core as core
+import radcad.wrappers as wrappers
 from radcad.utils import flatten, extract_exceptions
 
 import multiprocessing
@@ -10,6 +11,7 @@ import dill
 import ray
 
 from enum import Enum
+import copy
 
 
 cpu_count = multiprocessing.cpu_count() - 1 or 1
@@ -30,6 +32,7 @@ class Engine:
         self.processes = kwargs.pop("processes", cpu_count)
         self.backend = kwargs.pop("backend", Backend.DEFAULT)
         self.raise_exceptions = kwargs.pop("raise_exceptions", True)
+        self.deepcopy = kwargs.pop("deepcopy", True)
 
     def _run(self, experiment=None, **kwargs):
         if not experiment:
@@ -52,9 +55,12 @@ class Engine:
             )
             for sim in simulations
         ]
+
         result = []
 
-        self.experiment._before_experiment(engine=self)
+        self.experiment._before_experiment(experiment=self.experiment)
+
+        run_generator = self._run_stream(configs)
 
         if self.backend in [Backend.RAY, Backend.RAY_REMOTE]:
             if self.backend == Backend.RAY_REMOTE:
@@ -66,7 +72,7 @@ class Engine:
 
             futures = [
                 Engine._proxy_single_run_ray.remote((config, self.raise_exceptions))
-                for config in self._run_stream(configs)
+                for config in run_generator
             ]
             result = ray.get(futures)
         elif self.backend in [Backend.PATHOS, Backend.DEFAULT]:
@@ -75,7 +81,7 @@ class Engine:
                     Engine._proxy_single_run,
                     [
                         (config, self.raise_exceptions)
-                        for config in self._run_stream(configs)
+                        for config in run_generator
                     ],
                 )
         elif self.backend in [Backend.MULTIPROCESSING]:
@@ -86,35 +92,43 @@ class Engine:
                     Engine._proxy_single_run,
                     [
                         (config, self.raise_exceptions)
-                        for config in self._run_stream(configs)
+                        for config in run_generator
                     ],
                 )
         elif self.backend in [Backend.SINGLE_PROCESS]:
             result = [
                 Engine._proxy_single_run((config, self.raise_exceptions))
-                for config in self._run_stream(configs)
+                for config in run_generator
             ]
         else:
             raise Exception(f"Execution backend must be one of {Backend._member_names_}, not {self.backend}")
-
-        self.experiment._after_experiment(engine=self)
+        
         self.experiment.results, self.experiment.exceptions = extract_exceptions(result)
+        self.experiment._after_experiment(experiment=self.experiment)
         return self.experiment.results
 
     @ray.remote
     def _proxy_single_run_ray(args):
-        return Engine._single_run(args)
-
+        return Engine._proxy_single_run(args)
+    
     def _proxy_single_run(args):
         return Engine._single_run(args)
 
     def _single_run(args):
         run_args, raise_exceptions = args
         try:
-            results, exception = core.single_run(*run_args)
+            results, exception = core.single_run(*tuple(run_args))
             if raise_exceptions and exception:
                 raise exception
-            return results, exception
+            return results, {
+                    'exception': exception,
+                    'simulation': run_args.simulation,
+                    'run': run_args.run,
+                    'subset': run_args.subset,
+                    'timesteps': run_args.timesteps,
+                    'parameters': run_args.parameters,
+                    'initial_state': run_args.initial_state,
+                }
         except Exception as e:
             if raise_exceptions:
                 raise e
@@ -123,15 +137,17 @@ class Engine:
 
     def _get_simulation_from_config(config):
         states, state_update_blocks, params, timesteps, runs = config
-        model = core.Model(
+        model = wrappers.Model(
             initial_state=states, state_update_blocks=state_update_blocks, params=params
         )
-        return core.Simulation(model=model, timesteps=timesteps, runs=runs)
+        return wrappers.Simulation(model=model, timesteps=timesteps, runs=runs)
 
     def _run_stream(self, configs):
         simulations = [Engine._get_simulation_from_config(config) for config in configs]
 
         for simulation_index, simulation in enumerate(simulations):
+            simulation.index = simulation_index
+            
             timesteps = simulation.timesteps
             runs = simulation.runs
             initial_state = simulation.model.initial_state
@@ -140,34 +156,40 @@ class Engine:
             param_sweep = core.generate_parameter_sweep(params)
 
             self.experiment._before_simulation(
-                simulation=simulation, simulation_index=simulation_index
+                simulation=simulation
             )
 
             for run_index in range(0, runs):
-                self.experiment._before_run(simulation=simulation, run_index=run_index)
                 if param_sweep:
                     for subset, param_set in enumerate(param_sweep):
-                        yield (
+                        run_args = wrappers.Run(
                             simulation_index,
                             timesteps,
                             run_index,
                             subset,
-                            initial_state,
+                            copy.deepcopy(initial_state),
                             state_update_blocks,
-                            param_set,
+                            copy.deepcopy(param_set),
+                            self.deepcopy
                         )
+                        self.experiment._before_run(run=run_args)
+                        yield run_args
                 else:
-                    yield (
+                    run_args = wrappers.Run(
                         simulation_index,
                         timesteps,
                         run_index,
                         0,
-                        initial_state,
+                        copy.deepcopy(initial_state),
                         state_update_blocks,
-                        params,
+                        copy.deepcopy(params),
+                        self.deepcopy
                     )
-                self.experiment._after_run(simulation=simulation, run_index=run_index)
+                    self.experiment._before_run(run=run_args)
+                    yield run_args
+
+                self.experiment._after_run(run=run_args)
 
             self.experiment._after_simulation(
-                simulation=simulation, simulation_index=simulation_index
+                simulation=simulation
             )
